# Copyright 2013-2014 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from mock import patch, call
from unittest import TestCase
from itertools import product

from kamaki.clients import ClientError, astakos


example = dict(
    access=dict(
         token=dict(
            expires="2013-07-14T10:07:42.481134+00:00",
            id="ast@k0sT0k3n==",
            tenant=dict(
                id="42",
                name="Simple Name 0")
        ),
        serviceCatalog=[
            dict(name='service name 1', type='compute', endpoints=[
                dict(versionId='v1', publicUrl='http://1.1.1.1/v1'),
                dict(versionId='v2', publicUrl='http://1.1.1.1/v2')]),
            dict(name='service name 3', type='object-storage', endpoints=[
                dict(versionId='v2', publicUrl='http://1.1.1.1/v2'),
                dict(versionId='v2.1', publicUrl='http://1.1.1.1/v2/xtra')])
            ],
        user=dict(
            name='Simple Name 0',
            username='User Full Name 0',
            auth_token_expires='1362585796000',
            auth_token_created='1359931796000',
            email=['user0@example.gr'],
            id=42,
            uuid='aus3r-uu1d-507-73s71ng-as7ak0s')
        )
    )


class FR(object):
    json = example
    headers = {}
    content = json
    status = None
    status_code = 200

astakos_pkg = 'kamaki.clients.astakos'


class AstakosClient(TestCase):
    """Test synnefo.AstakosClient wrapper"""

    def setUp(self):
        self.url, self.token = 'https://astakos.example.com', 'ast@k0sT0k3n=='
        self.client = astakos.AstakosClient(self.url, self.token)

    @patch('astakosclient.AstakosClient.__init__')
    def test___init__(self, init):
        for args, kwargs in (
                (['some url', 'some token'], {}),
                (['some url', 'some token', 'other', 'params'], {}),
                (['some url', 'some token', 'other params'], (dict(k='v'))),
                (['some url', 'some token'], (dict(k1='v1', k2='v2'))),
                (['some url', ], (dict(k1='v1', k2='v2', token='some token'))),
            ):
            astakos.AstakosClient(*args, **kwargs)
            url, token = args.pop(0), kwargs.pop('token', None) or args.pop(0)
            self.assertTrue(
                init.mock_calls[-1], call(token, url, *args, **kwargs))

    @patch('%s.AstakosClient.get_endpoints' % astakos_pkg, return_value='ges')
    @patch('kamaki.clients.astakos.parse_endpoints', return_value=[
        dict(endpoints=['e1', 'e2', 'e3']), 'stuff'])
    def test_get_service_endpoints(self, parse_endpoints, get_endpoints):
        self.assertEqual(
            'e1', self.client.get_service_endpoints('service_type', 'version'))
        get_endpoints.assert_called_once()
        parse_endpoints.assert_called_once_with(
            'ges', ep_type='service_type', ep_version_id='version')

    @patch(
        '%s.AstakosClient.get_service_endpoints' % astakos_pkg,
        return_value=dict(publicURL='gse', itsnot='important'))
    def test_get_endpoint_url(self, get_service_endpoints):
        self.assertEqual(
            'gse', self.client.get_endpoint_url('serv_type', 'version'))
        get_service_endpoints.assert_called_once_with('serv_type', 'version')

    @patch('%s.AstakosClient.authenticate' % astakos_pkg, return_value=dict(
        access=dict(user='some user', itsnot='important'), isit='nope'))
    def test_user_info(self, authenticate):
        self.assertEqual('some user', self.client.user_info)
        authenticate.assert_called_once_with()

    @patch('%s.AstakosClient.authenticate' % astakos_pkg, return_value=dict(
        access=dict(user=dict(name='user'), itsnot='important'), isit='nope'))
    def test_user_term(self, authenticate):
        self.assertEqual('user', self.client.user_term('name'))
        authenticate.assert_called_once_with()


class LoggedAstakosClient(TestCase):
    """Test LoggedAstakosClient methods"""

    def setUp(self):
        self.url, self.token = 'https://astakos.example.com', 'ast@k0sT0k3n=='
        self.client = astakos.LoggedAstakosClient(self.url, self.token)

    def tearDown(self):
        FR.headers = {}

    @patch('kamaki.clients.recvlog.info', return_value='recvlog info')
    def test__dump_response(self, recvlog_info):
        for headers, status, message, data, LOG_DATA, LOG_TOKEN in product(
                (
                    {'k': 'v'},
                    {'X-Auth-Token': 'xxx'},
                    {'X-Auth-Token': 'xxx', 'k': 'v'}),
                (42, 'status'), ('message', ), ('data', 'My token is xxx'),
                (True, None), (True, None)):
            FR.headers = headers
            self.client.LOG_DATA, self.client.LOG_TOKEN = LOG_DATA, LOG_TOKEN
            if isinstance(status, int):
                self.client._dump_response(FR(), status, message, data)
                mock_calls = list(recvlog_info.mock_calls[-5:])
                self.assertEqual(
                    mock_calls.pop(),
                    call('-             -        -     -   -  - -'))
                size = len(data)
                if LOG_DATA:
                    token = headers.get('X-Auth-Token', '')
                    if token and not LOG_TOKEN:
                        data = data.replace(token, '...')
                    self.assertEqual(mock_calls.pop(), call(data))
                self.assertEqual(mock_calls[-2:], [
                    call('\n%s %s' % (status, message)),
                    call('data size: %s' % size)])
            else:
                self.assertRaises(
                    TypeError,
                    self.client._dump_response, FR(), status, message, data)

    @patch('%s.AstakosClient._call_astakos' % astakos_pkg, return_value='ret')
    def test__call_astakos(self, super_call):
        self.assertEqual(self.client._call_astakos('x', y='y'), 'ret')
        super_call.assert_called_once_with('x', y='y')


class CachedAstakosClient(TestCase):
    """Test Chached Astakos Client"""
    cached = False

    def assert_dicts_are_equal(self, d1, d2):
        for k, v in d1.items():
            self.assertTrue(k in d2)
            if isinstance(v, dict):
                self.assert_dicts_are_equal(v, d2[k])
            else:
                self.assertEqual(unicode(v), unicode(d2[k]))

    def setUp(self):
        self.url = 'https://astakos.example.com'
        self.token = 'ast@k0sT0k3n=='
        self.client = astakos.CachedAstakosClient(self.url, self.token)

    def tearDown(self):
        FR.json = example

    @patch('kamaki.clients.Client.__init__')
    def test___init__(self, super_init):
        c = astakos.CachedAstakosClient(self.url, self.token)
        super_init.assert_called_once_with(self.url, self.token)
        self.assertEqual(c._astakos, dict())
        self.assertEqual(c._uuids, dict())
        self.assertEqual(c._cache, dict())
        self.assertEqual(c._uuids2usernames, dict())
        self.assertEqual(c._usernames2uuids, dict())

    def test__resolve_token(self):
        for tok, exp in (
                (None, self.token), ('some token', 'some token'),
                (['token 1', 'token 2', 'token 3'], 'token 1')):
            self.assertEqual(exp, self.client._resolve_token(tok))
        self.client.token = None
        self.assertRaises(AssertionError, self.client._resolve_token, None)

    @patch(
        '%s.CachedAstakosClient._resolve_token' % astakos_pkg,
        return_value='rtoken')
    @patch('%s.CachedAstakosClient._validate_token' % astakos_pkg)
    def test_get_client(self, validate, resolve):
        self.client._uuids['rtoken'] = 'ruuid'
        self.client._astakos['ruuid'] = 'rclient'
        self.assertEqual('rclient', self.client.get_client('not important'))
        validate.assert_called_once_with('rtoken')
        resolve.assert_called_once_with('not important')

    @patch(
        '%s.CachedAstakosClient._resolve_token' % astakos_pkg,
        return_value='rtoken')
    @patch('%s.LoggedAstakosClient.__init__' % astakos_pkg, return_value=None)
    @patch(
        '%s.LoggedAstakosClient.authenticate' % astakos_pkg,
        return_value=example)
    def test_authenticate(self, authenticate, super_init, resolve):
        self.assertEqual(example, self.client.authenticate('not important'))
        resolve.assert_called_once_with('not important')
        super_init.assert_called_once_with(
            self.url, 'rtoken', logger=astakos.log)
        authenticate.assert_called_once_with()
        uuid = example['access']['user']['id']
        self.assertEqual(self.client._uuids['rtoken'], uuid)
        self.assertEqual(self.client._cache[uuid], example)
        self.assertTrue(isinstance(
            self.client._astakos[uuid], astakos.LoggedAstakosClient))
        self.assertFalse(self.client._astakos[uuid].LOG_TOKEN)
        self.assertFalse(self.client._astakos[uuid].LOG_DATA)
        self.assertEqual(self.client._uuids2usernames[uuid], dict())
        self.assertEqual(self.client._usernames2uuids[uuid], dict())

        self.client.LOG_TOKEN, self.client.LOG_DATA = 'tkn', 'dt'
        self.client.authenticate()
        self.assertEqual(resolve.mock_calls[-1], call(None))
        self.assertEqual(self.client._astakos[uuid].LOG_TOKEN, 'tkn')
        self.assertEqual(self.client._astakos[uuid].LOG_DATA, 'dt')

    @patch(
        '%s.CachedAstakosClient.get_token' % astakos_pkg, return_value='t1')
    def test_remove_user(self, get_token):
        self.client._uuids = dict(t1='v1', t2='v2')
        self.client._cache = dict(u1='v1', u2='v2')
        self.client._astakos = dict(u1='v1', u2='v2')
        self.client._uuids2usernames = dict(u1='v1', u2='v2')
        self.client._usernames2uuids = dict(u1='v1', u2='v2')
        self.client.remove_user('u1')
        get_token.assert_called_once_with('u1')
        self.assertEqual(self.client._uuids, dict(t2='v2'))
        self.assertEqual(self.client._cache, dict(u2='v2'))
        self.assertEqual(self.client._astakos, dict(u2='v2'))
        self.assertEqual(self.client._uuids2usernames, dict(u2='v2'))
        self.assertEqual(self.client._usernames2uuids, dict(u2='v2'))
        self.assertRaises(KeyError, self.client.remove_user, 'u1')

    def test_get_token(self):
        token = example['access']['token']['id']
        self.client._cache['uuid'] = example
        self.assertEqual(self.client.get_token('uuid'), token)
        self.assertRaises(KeyError, self.client.get_token, 'non uuid')

    @patch('%s.CachedAstakosClient.get_token' % astakos_pkg, return_value='t1')
    @patch('%s.CachedAstakosClient.authenticate' % astakos_pkg)
    def test__validate_token(self, authenticate, get_token):
        self.client._uuids['t1'] = 'u1'
        self.client._validate_token('t1')
        self.assertEqual(get_token.mock_calls[-1], call('u1'))
        self.assertEqual(authenticate.mock_calls, [])
        self.assertTrue('t1' in self.client._uuids)

        self.client._uuids['t2'] = 'u2'
        self.client._validate_token('t2')
        self.assertEqual(get_token.mock_calls[-1], call('u2'))
        self.assertEqual(authenticate.mock_calls[-1], call('t2'))
        self.assertTrue('t2' not in self.client._uuids)

        self.client._validate_token('t3')
        self.assertEqual(authenticate.mock_calls[-1], call('t3'))

    @patch(
        '%s.CachedAstakosClient._resolve_token' % astakos_pkg,
        return_value='tkn')
    @patch('%s.CachedAstakosClient._validate_token' % astakos_pkg)
    def test_get_services(self, validate, resolve):
        self.client._cache['u1'], self.client._uuids['tkn'] = example, 'u1'
        self.assertEqual(
            self.client.get_services('dont care'),
            example['access']['serviceCatalog'])
        resolve.assert_called_once_with('dont care')
        validate.assert_called_once_with('tkn')

    @patch(
        '%s.CachedAstakosClient.get_services' % astakos_pkg,
        return_value=example['access']['serviceCatalog'])
    def test_get_service_details(self, get_services):
        self.assertEqual(
            example['access']['serviceCatalog'][0],
            self.client.get_service_details('compute', 'dont care'))
        get_services.assert_called_once_with('dont care')
        self.assertRaises(
            astakos.AstakosClientError, self.client.get_service_details,
            'non-existing type', 'dont care')

    @patch(
        '%s.CachedAstakosClient.get_service_details' % astakos_pkg,
        return_value=example['access']['serviceCatalog'][0])
    def test_get_service_endpoints(self, get_service_details):
        service = example['access']['serviceCatalog'][0]
        self.assertEqual(
            self.client.get_service_endpoints('not important', 'v1'),
            service['endpoints'][0])
        get_service_details.assert_called_once_with('not important', None)
        self.assertRaises(
            astakos.AstakosClientError, self.client.get_service_endpoints,
            'dont care', 'vX')

    @patch(
        '%s.CachedAstakosClient.get_service_endpoints' % astakos_pkg,
        return_value=dict(publicURL='a URL'))
    def test_get_endpoint_url(self, get_service_endpoints):
        args = ('stype', 'version', 'token')
        self.assertEqual(self.client.get_endpoint_url(*args), 'a URL')
        get_service_endpoints.assert_called_once_with(*args)

    @patch('%s.CachedAstakosClient.authenticate' % astakos_pkg)
    @patch('%s.CachedAstakosClient.get_token' % astakos_pkg, return_value='t1')
    def test_list_users(self, get_token, authenticate):
        self.client._cache = dict()
        self.assertEqual(self.client.list_users(), [])

        e1 = dict(access=dict(
            token=dict(id='t1', otherstuff='...'),
            user=dict(name='user 1', otherstuff2='...')))
        e2 = dict(access=dict(
            token=dict(id='t2', otherstuff='...'),
            user=dict(name='user 2', otherstuff2='...')))
        self.client._cache = dict(u1=e1, u2=e2)
        e1['access']['user']['auth_token'] = 't1'
        e2['access']['user']['auth_token'] = 't1'
        self.assertEqual(
            sorted(self.client.list_users()),
            sorted([e1['access']['user'], e2['access']['user']]))

    @patch(
        '%s.CachedAstakosClient._resolve_token' % astakos_pkg,
        return_value='t1')
    @patch('%s.CachedAstakosClient._validate_token' % astakos_pkg)
    def test_user_info(self, validate, resolve):
        self.client._uuids = dict(t1='u1', t2='u2')
        self.client._cache = dict(u1=example, u2='nothing')
        self.assertEqual(
            self.client.user_info('dont care'), example['access']['user'])
        resolve.assert_called_once_with('dont care')
        validate.assert_called_once_with('t1')

    @patch('%s.CachedAstakosClient.user_info' % astakos_pkg, return_value=dict(
        key='val'))
    def test_user_term(self, user_info):
        self.assertEqual('val', self.client.user_term('key', 'dont care'))
        user_info.assert_called_once_with('dont care')
        self.assertEqual(None, self.client.user_term('not key', 'dont care'))

    @patch('%s.CachedAstakosClient.user_term' % astakos_pkg, return_value='rt')
    def test_term(self, user_term):
        self.assertEqual('rt', self.client.term('key', 'dont care'))
        user_term.assert_called_once_with('key', 'dont care')

    @patch(
        '%s.CachedAstakosClient.uuids2usernames' % astakos_pkg,
        return_value='a user name list')
    @patch(
        '%s.CachedAstakosClient.usernames2uuids' % astakos_pkg,
        return_value='a uuid list')
    def test_post_user_catalogs(self, usernames2uuids, uuids2usernames):
        self.assertEqual('a user name list', self.client.post_user_catalogs(
            uuids=['u1'], token='X'))
        self.assertEqual('a uuid list', self.client.post_user_catalogs(
            displaynames=['n1', 'n2'], token='X'))
        uuids2usernames.assert_called_once_with(['u1'], 'X')
        usernames2uuids.assert_called_once_with(['n1', 'n2'], 'X')

    @patch(
        'astakosclient.AstakosClient.get_usernames',
        return_value=dict(uuid1='name 1', uuid2='name 2'))
    @patch(
        '%s.CachedAstakosClient._resolve_token' % astakos_pkg,
        return_value='t1')
    @patch('%s.CachedAstakosClient._validate_token' % astakos_pkg)
    @patch('astakosclient.AstakosClient.__init__', return_value=None)
    def test_uuids2usernames(
            self, orig_astakos, validate, resolve, get_usernames):
        import astakosclient
        self.client._uuids['t1'] = 'uuid0'
        self.client._astakos['uuid0'] = astakosclient.AstakosClient(
            self.url, self.token)
        self.client._uuids2usernames['uuid0'] = dict(uuid0='name 0')
        exp = dict()
        for i in range(3):
            exp['uuid%s' % i] = 'name %s' % i
        self.assertEqual(exp, self.client.uuids2usernames(
            ['uuid1', 'uuid2'], 'dont care'))
        resolve.assert_called_once_with('dont care')
        validate.assert_called_once_with('t1')
        get_usernames.assert_called_once_with(['uuid1', 'uuid2'])

    @patch(
        'astakosclient.AstakosClient.get_uuids',
        return_value=dict(name1='uuid 1', name2='uuid 2'))
    @patch(
        '%s.CachedAstakosClient._resolve_token' % astakos_pkg,
        return_value='t1')
    @patch('%s.CachedAstakosClient._validate_token' % astakos_pkg)
    @patch('astakosclient.AstakosClient.__init__', return_value=None)
    def test_usernames2uuids(
            self, orig_astakos, validate, resolve, get_uuids):
        import astakosclient
        self.client._uuids['t1'] = 'uuid 0'
        self.client._astakos['uuid 0'] = astakosclient.AstakosClient(
            self.url, self.token)
        self.client._usernames2uuids['uuid 0'] = dict(name0='uuid 0')
        exp = dict()
        for i in range(3):
            exp['name%s' % i] = 'uuid %s' % i
        self.assertEqual(exp, self.client.usernames2uuids(
            ['name1', 'name2'], 'dont care'))
        resolve.assert_called_once_with('dont care')
        validate.assert_called_once_with('t1')
        get_uuids.assert_called_once_with(['name1', 'name2'])


if __name__ == '__main__':
    from sys import argv
    from kamaki.clients.test import runTestCase
    not_found = True
    if not argv[1:] or argv[1] == 'AstakosClient':
        not_found = False
        runTestCase(AstakosClient, 'Kamaki Astakos Client', argv[2:])
    if not argv[1:] or argv[1] == 'LoggedAstakosClient':
        not_found = False
        runTestCase(LoggedAstakosClient, 'Logged Astakos Client', argv[2:])
    if not argv[1:] or argv[1] == 'CachedAstakosClient':
        not_found = False
        runTestCase(CachedAstakosClient, 'Cached Astakos Client', argv[2:])
    if not_found:
        print('TestCase %s not found' % argv[1])



# Copyright 2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from kamaki.clients import ClientError, Waiter
from kamaki.clients.network.rest_api import NetworkRestClient


class NetworkClient(NetworkRestClient, Waiter):
    """OpenStack Network API 2.0 client"""

    def list_networks(self):
        r = self.networks_get(success=200)
        return r.json['networks']

    def create_network(self, name, admin_state_up=None, shared=None):
        req = dict(network=dict(
            name=name, admin_state_up=bool(admin_state_up)))
        if shared not in (None, ):
            req['network']['shared'] = bool(shared)
        r = self.networks_post(json_data=req, success=201)
        return r.json['network']

    def create_networks(self, networks):
        """Atomic operation for batch network creation (all or nothing)

        :param networks: (list) [
            {name: ..(str).., admin_state_up: ..(bool).., shared: ..(bool)..},
            {name: ..(str).., admin_state_up: ..(bool).., shared: ..(bool)..}]
            name is mandatory, the rest is optional
            e.g., create_networks([
            {name: 'net1', admin_state_up: True},
            {name: 'net2'}])
        :returns: (list of dicts) created networks details
        :raises ValueError: if networks is misformated
        :raises ClientError: if the request failed or didn't return 201
        """
        try:
            msg = 'The networks parameter must be list or tuple'
            assert (
                isinstance(networks, list) or isinstance(networks, tuple)), msg
            for network in networks:
                msg = 'Network specification %s is not a dict' % network
                assert isinstance(network, dict), msg
                err = set(network).difference(
                    ('name', 'admin_state_up', 'shared'))
                if err:
                    raise ValueError(
                        'Invalid key(s): %s in network specification %s' % (
                            err, network))
                msg = 'Name is missing in network specification: %s' % network
                assert network.get('name', None), msg
                network.setdefault('admin_state_up', False)
        except AssertionError as ae:
            raise ValueError('%s' % ae)

        req = dict(networks=list(networks))
        r = self.networks_post(json_data=req, success=201)
        return r.json['networks']

    def get_network_details(self, network_id):
        r = self.networks_get(network_id, success=200)
        return r.json['network']

    def update_network(
            self, network_id, name=None, admin_state_up=None, shared=None):
        network = dict()
        if name:
            network['name'] = name
        if admin_state_up not in (None, ):
            network['admin_state_up'] = admin_state_up
        if shared not in (None, ):
            network['shared'] = shared
        network = dict(network=network)
        r = self.networks_put(network_id, json_data=network, success=200)
        return r.json['network']

    def delete_network(self, network_id):
        r = self.networks_delete(network_id, success=204)
        return r.headers

    def list_subnets(self):
        r = self.subnets_get(success=200)
        return r.json['subnets']

    def create_subnet(
            self, network_id, cidr,
            name=None, allocation_pools=None, gateway_ip=None, subnet_id=None,
            ipv6=None, enable_dhcp=None):
        """
        :param network_id: (str)
        :param cidr: (str)

        :param name: (str) The subnet name
        :param allocation_pools: (list of dicts) start/end addresses of
            allocation pools: [{'start': ..., 'end': ...}, ...]
        :param gateway_ip: (str)
        :param subnet_id: (str)
        :param ipv6: (bool) ip_version == 6 if true else 4 (default)
        :param enable_dhcp: (bool)
        """
        subnet = dict(
            network_id=network_id, cidr=cidr, ip_version=6 if ipv6 else 4)
        if name:
            subnet['name'] = name
        if allocation_pools:
            subnet['allocation_pools'] = allocation_pools
        if gateway_ip:
            subnet['gateway_ip'] = gateway_ip
        if subnet_id:
            subnet['id'] = subnet_id
        if enable_dhcp not in (None, ):
            subnet['enable_dhcp'] = bool(enable_dhcp)
        r = self.subnets_post(json_data=dict(subnet=subnet), success=201)
        return r.json['subnet']

    def create_subnets(self, subnets):
        """Atomic operation for batch subnet creation (all or nothing)

        :param subnets: (list of dicts) {key: ...} with all parameters in the
            method create_subnet, where method mandatory / optional paramteres
            respond to mandatory / optional paramters in subnets items
        :returns: (list of dicts) created subnetss details
        :raises ValueError: if subnets parameter is incorrectly formated
        :raises ClientError: if the request failed or didn't return 201
        """
        try:
            msg = 'The subnets parameter must be list or tuple'
            assert (
                isinstance(subnets, list) or isinstance(subnets, tuple)), msg
            for subnet in subnets:
                msg = 'Subnet specification %s is not a dict' % subnet
                assert isinstance(subnet, dict), msg
                err = set(subnet).difference((
                    'network_id', 'cidr', 'name', 'allocation_pools',
                    'gateway_ip', 'subnet_id', 'ipv6', 'enable_dhcp'))
                if err:
                    raise ValueError(
                        'Invalid key(s): %s in subnet specification %s' % (
                            err, subnet))
                msg = 'network_id is missing in subnet spec: %s' % subnet
                assert subnet.get('network_id', None), msg
                msg = 'cidr is missing in subnet spec: %s' % subnet
                assert subnet.get('cidr', None), msg
                subnet['ip_version'] = 6 if subnet.pop('ipv6', None) else 4
                if 'subnet_id' in subnet:
                    subnet['id'] = subnet.pop('subnet_id')
        except AssertionError as ae:
            raise ValueError('%s' % ae)

        r = self.subnets_post(
            json_data=dict(subnets=list(subnets)), success=201)
        return r.json['subnets']

    def get_subnet_details(self, subnet_id):
        r = self.subnets_get(subnet_id, success=200)
        return r.json

    def update_subnet(
            self, subnet_id,
            name=None, allocation_pools=None, gateway_ip=None, ipv6=None,
            enable_dhcp=None):
        """
        :param subnet_id: (str)

        :param name: (str) The subnet name
        :param allocation_pools: (list of dicts) start/end addresses of
            allocation pools: [{'start': ..., 'end': ...}, ...]
        :param gateway_ip: (str)
        :param ipv6: (bool) ip_version == 6 if true, 4 if false, used as filter
        :param enable_dhcp: (bool)
        """
        subnet = dict()
        if name not in (None, ):
            subnet['name'] = name
        if allocation_pools not in (None, ):
            subnet['allocation_pools'] = allocation_pools
        if gateway_ip not in (None, ):
            subnet['gateway_ip'] = gateway_ip
        if ipv6 not in (None, ):
            subnet['ip_version'] = 6 if ipv6 else 4
        if enable_dhcp not in (None, ):
            subnet['enable_dhcp'] = enable_dhcp
        r = self.subnets_put(
            subnet_id, json=dict(subnet=subnet), success=200)
        return r.json['subnet']

    def delete_subnet(self, subnet_id):
        r = self.subnets_delete(subnet_id, success=204)
        return r.headers

    def list_ports(self):
        r = self.ports_get(success=200)
        return r.json['ports']

    def create_port(
            self, network_id,
            name=None, status=None, admin_state_up=None, mac_address=None,
            fixed_ips=None, security_groups=None):
        """
        :param network_id: (str)

        :param name: (str)
        :param status: (str)
        :param admin_state_up: (bool) Router administrative status (UP / DOWN)
        :param mac_address: (str)
        :param fixed_ips: (str)
        :param security_groups: (list)
        """
        port = dict(network_id=network_id)
        if name:
            port['name'] = name
        if status:
            port['status'] = status
        if admin_state_up not in (None, ):
            port['admin_state_up'] = bool(admin_state_up)
        if mac_address:
            port['mac_address'] = mac_address
        if fixed_ips:
            port['fixed_ips'] = fixed_ips
        if security_groups:
            port['security_groups'] = security_groups
        r = self.ports_post(json_data=dict(port=port), success=201)
        return r.json['port']

    def create_ports(self, ports):
        """Atomic operation for batch port creation (all or nothing)

        :param ports: (list of dicts) {key: ...} with all parameters in the
            method create_port, where method mandatory / optional paramteres
            respond to mandatory / optional paramters in ports items
        :returns: (list of dicts) created portss details
        :raises ValueError: if ports parameter is incorrectly formated
        :raises ClientError: if the request failed or didn't return 201
        """
        try:
            msg = 'The ports parameter must be list or tuple'
            assert (
                isinstance(ports, list) or isinstance(ports, tuple)), msg
            for port in ports:
                msg = 'Subnet specification %s is not a dict' % port
                assert isinstance(port, dict), msg
                err = set(port).difference((
                    'network_id', 'status', 'name', 'admin_state_up',
                    'mac_address', 'fixed_ips', 'security_groups'))
                if err:
                    raise ValueError(
                        'Invalid key(s): %s in port specification %s' % (
                            err, port))
                msg = 'network_id is missing in port spec: %s' % port
                assert port.get('network_id', None), msg
        except AssertionError as ae:
            raise ValueError('%s' % ae)
        r = self.ports_post(json_data=dict(ports=list(ports)), success=201)
        return r.json['ports']

    def get_port_details(self, port_id):
        r = self.ports_get(port_id, success=200)
        return r.json['port']

    def delete_port(self, port_id):
        r = self.ports_delete(port_id, success=204)
        return r.headers

    def update_port(
            self, port_id, network_id,
            name=None, status=None, admin_state_up=None, mac_address=None,
            fixed_ips=None, security_groups=None):
        """
        :param network_id: (str)

        :param name: (str)
        :param status: (str)
        :param admin_state_up: (bool) Router administrative status (UP / DOWN)
        :param mac_address: (str)
        :param fixed_ips: (str)
        :param security_groups: (list)
        """
        port = dict(network_id=network_id)
        if name:
            port['name'] = name
        if status:
            port['status'] = status
        if admin_state_up not in (None, ):
            port['admin_state_up'] = bool(admin_state_up)
        if mac_address:
            port['mac_address'] = mac_address
        if fixed_ips:
            port['fixed_ips'] = fixed_ips
        if security_groups:
            port['security_groups'] = security_groups
        r = self.ports_put(port_id, json_data=dict(port=port), success=200)
        return r.json['port']

    def list_floatingips(self):
        r = self.floatingips_get(success=200)
        return r.json['floatingips']

    def get_floatingip_details(self, floatingip_id):
        r = self.floatingips_get(floatingip_id, success=200)
        return r.json['floatingip']

    def create_floatingip(
            self, floating_network_id,
            floating_ip_address='', port_id='', fixed_ip_address=''):
        floatingip = dict(floating_network_id=floating_network_id)
        if floating_ip_address:
            floatingip['floating_ip_address'] = floating_ip_address
        if port_id:
            floatingip['port_id'] = port_id
        if fixed_ip_address:
            floatingip['fixed_ip_address'] = fixed_ip_address
        r = self.floatingips_post(
            json_data=dict(floatingip=floatingip), success=200)
        return r.json['floatingip']

    def update_floatingip(
            self, floatingip_id,
            floating_ip_address='', port_id='', fixed_ip_address=''):
        """To nullify something optional, use None"""
        floatingip = dict()
        if floating_ip_address != '':
            floatingip['floating_ip_address'] = floating_ip_address
        if port_id != '':
            floatingip['port_id'] = port_id
        if fixed_ip_address != '':
            floatingip['fixed_ip_address'] = fixed_ip_address
        r = self.floatingips_put(
            floatingip_id, json_data=dict(floatingip=floatingip), success=200)
        return r.json['floatingip']

    def delete_floatingip(self, floatingip_id):
        r = self.floatingips_delete(floatingip_id, success=204)
        return r.headers

    #  Wait methods

    def wait_port(
            self, port_id,
            current_status='BUILD', delay=1, max_wait=100, wait_cb=None):

        def get_status(self, net_id):
            r = self.get_port_details(port_id)
            return r['status'], None

        return self._wait(
            port_id, current_status, get_status, delay, max_wait, wait_cb)



# Copyright 2011-2014 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from kamaki.clients import Client, ClientError, quote
from kamaki.clients.utils import path4url


def _format_image_headers(headers):
    reply = dict(properties=dict())
    meta_prefix = 'x-image-meta-'
    property_prefix = 'x-image-meta-property-'

    for key, val in headers.items():
        key = key.lower()
        if key.startswith(property_prefix):
            key = key[len(property_prefix):].upper().replace('-', '_')
            reply['properties'][key] = val
        elif key.startswith(meta_prefix):
            key = key[len(meta_prefix):]
            reply[key] = val
    return reply


class ImageClient(Client):
    """Synnefo Plankton API client"""
    service_type = 'image'

    def __init__(self, endpoint_url, token):
        super(ImageClient, self).__init__(endpoint_url, token)
        self.request_headers_to_quote = ['X-Image-Meta-Name', ]
        self.request_header_prefices_to_quote = ['X-Image-Meta-Property-', ]
        self.response_headers = [
            'X-Image-Meta-Name', 'X-Image-Meta-Location',
            'X-Image-Meta-Description']
        self.response_header_prefices = ['X-Image-Meta-Property-', ]

    def list_public(self, detail=False, filters={}, order=''):
        """
        :param detail: (bool)

        :param filters: (dict) request filters

        :param order: (str) order listing by field (default is ascending, - for
            descending)

        :returns: (list) id,name + full image info if detail
        """
        path = path4url('images', 'detail') if detail else (
            '%s/' % path4url('images'))

        async_params = {}
        if isinstance(filters, dict):
            for key, value in filters.items():
                if value:
                    async_params[key] = value
        if order and order.startswith('-'):
            async_params['sort_dir'] = 'desc'
            order = order[1:]
        else:
            async_params['sort_dir'] = 'asc'
        if order:
            async_params['sort_key'] = order

        r = self.get(path, async_params=async_params, success=200)
        return r.json

    def get_meta(self, image_id):
        """
        :param image_id: (string)

        :returns: (list) image metadata (key:val)
        """
        path = path4url('images', image_id)
        r = self.head(path, success=200)

        return _format_image_headers(r.headers)

    def register(self, name, location, params={}, properties={}):
        """Register an image that is uploaded at location

        :param name: (str)

        :param location: (str or iterable) if iterable, then
            (user_uuid, container, image_path) else if string
            pithos://<user_uuid>/<container>/<image object>

        :param params: (dict) image metadata (X-Image-Meta) can be id, store,
            disc_format, container_format, size, checksum, is_public, owner

        :param properties: (dict) image properties (X-Image-Meta-Property)

        :returns: (dict) metadata of the created image
        """
        path = '%s/' % path4url('images')
        self.set_header('X-Image-Meta-Name', name)
        location = location if (
            isinstance(location, str) or isinstance(location, unicode)) else (
                'pithos://%s' % '/'.join(location))
        prefix = 'pithos://'
        if location.startswith(prefix):
            lvalues = (location[len(prefix):]).split('/')
            location = '%s%s' % (prefix, '/'.join([
                quote(s.encode('utf-8')) for s in lvalues]))
        else:
            lvalues = location.split('/')
            location = '.'.join([quote(s.encode('utf-8')) for s in lvalues])
        self.set_header('X-Image-Meta-Location', location)

        async_headers = {}
        for key, val in params.items():
            if key in ('store', 'disk_format', 'container_format',
                       'size', 'checksum', 'is_public', 'owner') and val:
                key = 'x-image-meta-' + key.replace('_', '-')
                async_headers[key] = val

        for key, val in properties.items():
            async_headers['x-image-meta-property-%s' % key] = val

        r = self.post(path, success=200, async_headers=async_headers)

        return _format_image_headers(r.headers)

    def unregister(self, image_id):
        """Unregister an image

        :param image_id: (str)

        :returns: (dict) response headers
        """
        path = path4url('images', image_id)
        r = self.delete(path, success=204)
        return r.headers

    def list_members(self, image_id):
        """
        :param image_id: (str)

        :returns: (list) users who can use current user's images
        """
        path = path4url('images', image_id, 'members')
        r = self.get(path, success=200)
        return r.json['members']

    def list_shared(self, member):
        """
        :param member: (str) sharers account

        :returns: (list) images shared by member
        """
        path = path4url('shared-images', member)
        #self.set_param('format', 'json')
        r = self.get(path, success=200)
        return r.json['shared_images']

    def add_member(self, image_id, member):
        """
        :param image_id: (str)

        :param member: (str) user to allow access to current user's images
        """
        path = path4url('images', image_id, 'members', member)
        self.set_header('Content-Length', 0)
        r = self.put(path, success=204)
        return r.headers

    def remove_member(self, image_id, member):
        """
        :param image_id: (str)

        :param member: (str) user to deprive from current user's images
        """
        path = path4url('images', image_id, 'members', member)
        r = self.delete(path, success=204)
        return r.headers

    def set_members(self, image_id, members):
        """
        :param image_id: (str)

        :param members: (list) user to deprive from current user's images
        """
        path = path4url('images', image_id, 'members')
        req = {'memberships': [{'member_id': member} for member in members]}
        r = self.put(path, json=req, success=204)
        return r.headers

    def update_image(
            self, image_id,
            name=None, disk_format=None, container_format=None,
            status=None, public=None, owner_id=None, **properties):
        path = path4url('images', image_id)
        if name is not None:
            self.set_header('X-Image-Meta-Name', name)
        if disk_format is not None:
            self.set_header('X-Image-Meta-Disk-Format', disk_format)
        if container_format is not None:
            self.set_header('X-Image-Meta-Container-Format', container_format)
        if status is not None:
            self.set_header('X-Image-Meta-Status', status)
        if public is not None:
            self.set_header('X-Image-Meta-Is-Public', bool(public))
        if owner_id is not None:
            self.set_header('X-Image-Meta-Owner', owner_id)
        for k, v in properties.items():
            self.set_header('X-Image-Meta-Property-%s' % k, v)
        self.set_header('Content-Length', 0)
        r = self.put(path, success=200)
        return r.headers
